#ifndef VEC_H
#define VEC_H

#include <algorithm>
#include <cstddef>
#include <memory>

using std::max;

template <class T> class Vec {
public:
	typedef T* iterator;
	typedef const T* const_iterator;
	typedef size_t size_type;
	typedef T value_type;
	typedef T& reference;
	typedef const T& const_reference;

	Vec() { vytvoreni(); }
	explicit Vec(size_type n, const T& t = T()) { vytvoreni(n, t); }

	Vec(const Vec& v) { vytvoreni(v.begin(), v.end()); }
	Vec& operator=(const Vec&);	// definovno na strnce ###11.3.2/196###
	~Vec() { zruseni(); }

	T& operator[](size_type i) { return data[i]; }
	const T& operator[](size_type i) const { return data[i]; }

	void push_back(const T& t) {
		if (volny == limit)
			rozsireni();
		nekontrolovane_pripojeni(t);
	}

	size_type size() const { return volny - data; }  // zmnno

	iterator begin() { return data; }
	const_iterator begin() const { return data; }

	iterator end() { return volny; }                 // zmnno
	const_iterator end() const { return volny; }     // zmnno
	void clear() { zruseni(); }
	bool empty() const { return data == volny; }

private:
	iterator data;	// ukazuje na poten prvek objektu typu 'Vec'
	iterator volny;	// ukazuje na (jednu pozici za) posledn prvek objektu typu 'Vec'
	iterator limit;	// ukazuje na (jednu pozici za) alokovanou pam

	// sluby alokace pamti
	std::allocator<T> alloc;	// objekt zen alokace pamti

	// alokuje a inicializuje implementan pole
	void vytvoreni();
	void vytvoreni(size_type, const T&);
	void vytvoreni(const_iterator, const_iterator);

	// ru prvky v poli a uvoluje pam
	void zruseni();

	// funkce pro podporu 'push_back'
	void rozsireni();
	void nekontrolovane_pripojeni(const T&);
};

template <class T> void Vec<T>::vytvoreni()
{
	data = volny = limit = 0;
}

template <class T> void Vec<T>::vytvoreni(size_type n, const T& hod)
{
	data = alloc.allocate(n);
	limit = volny = data + n;
	std::uninitialized_fill(data, limit, hod);
}

template <class T>
void Vec<T>::vytvoreni(const_iterator i, const_iterator j)
{
	data = alloc.allocate(j - i);
	limit = volny = std::uninitialized_copy(i, j, data);
}

template <class T> void Vec<T>::zruseni()
{
	if (data) {
		// zru (v obrcenm poad) vytvoen prvky
		iterator it = volny;
		while (it != data)
			alloc.destroy(--it);

		// vrt systmu vechen alokovan prostor
		alloc.deallocate(data, limit - data);
	}
	// penastav ukazatele tak, aby znovu oznaovaly przdn objekt typu 'Vec'
	data = limit = volny = 0;
}

template <class T> void Vec<T>::rozsireni()
{
	// pi roziovn alokujeme dvakrt tolik msta, jako prv pouvme
	size_type nova_velikost = max(2 * (limit - data), ptrdiff_t(1));

	// alokuje nov prostor a zkopruje do nj existujc prvky
	iterator nova_data = alloc.allocate(nova_velikost);
	iterator novy_volny = std::uninitialized_copy(data, volny, nova_data);

	// vrt star prostor systmu
	zruseni();

	// nastav ukazatele tak, aby ukazovaly na nov alokovan prostor
	data = nova_data;
	volny = novy_volny;
	limit = data + nova_velikost;
}

// pedpokld, e 'volny' ukazuje na alokovan, ale neinicializovan prostor
template <class T> void Vec<T>::nekontrolovane_pripojeni(const T& hod)
{
	alloc.construct(volny++, hod);
}

template <class T>
Vec<T>& Vec<T>::operator=(const Vec& pso)
{
	// kontrola piazen sebe sama
	if (&pso != this) {

		// uvoln pole na lev stran opertoru
		zruseni();

		// kopruje prvky z prav strany opertoru na levou
		vytvoreni(pso.begin(), pso.end());
	}
	return *this;
}

#endif
